--					FUNCTIONS					--

--Acquires and validates the active sprite
local function acquire_sprite()
	local sprite = app.sprite
	if sprite == nil then
		app.alert{ title = "Error", text = "No sprite was selected" }
		return nil
	end
	if sprite.colorMode ~= ColorMode.INDEXED then
		app.alert{ title = "Error", text = "Sprite colours must be indexed" }
		return nil
	end
	if #sprite.palettes[1] ~= 4 then
		app.alert{ title = "Error", text = "Sprite palette must have 4 colours" }
		return
	end
	if sprite.width ~= 8 or (sprite.height ~= 8 and sprite.height ~= 16) then
		app.alert{ title = "Error", text = "Sprite size must be 8x8 or 8x16" }
		return nil
	end
	return sprite
end

--Creates and sets up the custom export dialog
local function create_dialog(path)
	local dialog = Dialog("GameBoy Pattern Exporter")
	local function set_assembly()
		dialog:modify{
			id = "header",
			visible = false
		}
		dialog:modify{
			id = "source",
			visible = false
		}
		dialog:modify{
			id = "asm",
			visible = true
		}
		dialog:modify{
			id = "location",
			text = "Select a location to save the assembly file."
		}
	end
	local function set_c()
		dialog:modify{
			id = "asm",
			visible = false
		}
		dialog:modify{
			id = "header",
			visible = true
		}
		dialog:modify{
			id = "source",
			visible = true
		}
		dialog:modify{
			id = "location",
			text = "Select a location to save the header and source files."
		}
	end
	dialog:label{ text = "Basic output details" }
	dialog:entry{
		id = "variable",
		label = "Variable",
		text = app.fs.fileName(path):gsub(app.fs.filePath(path), "")
	}
	dialog:radio{
		id = "assembly",
		label = "Language",
		text = "Assembly",
		selected = true,
		onclick = set_assembly
	}
	dialog:radio{
		id = "c",
		text = "C/C++",
		selected = false,
		onclick = set_c
	}
	dialog:radio{
		id = "binary",
		label = "Format",
		text = "Binary",
		selected = true
	}
	dialog:radio{
		id = "hexadecimal",
		text = "Hexadecimal",
		selected = false
	}
	dialog:separator{ }
	dialog:label{
		id = "location",
		text = "Select a location to save the assembly file."
	}
	dialog:file{
		id = "asm",
		label = "Output",
		title = "GameBoy-Assembly File Export",
		open = false,
		save = true,
		filename = path..".asm",
		filetypes = { "asm", "s" }
	}
	dialog:file{
		id = "header",
		label = "Header",
		title = "GameBoy-C Header Export",
		open = false,
		save = true,
		filename = path..".h",
		filetypes = { "h", "hpp" },
		visible = false
	}
	dialog:file{
		id = "source",
		label = "Source",
		Title = "GameBoy-C Source Export",
		open = false,
		save = true,
		filename = path..".c",
		filetypes = { "c", "cpp" },
		visible = false
	}
	dialog:separator{ }
	dialog:button{
		id = "confirm",
		text = "Confirm"
	}
	dialog:button{
		id = "cancel",
		text = "Cancel"
	}
	return dialog
end

--Outputs a string matching the given number and using the specified format (Binary or Hexadecimal)
local function format_chunk(number, hex)
	if hex then
		return "0x"..string.format("%02x", number);
	else
		local bits = { }
		while number > 0 do
			local current = math.fmod(number, 2)
			bits[#bits + 1] = current
			number = (number - current) / 2
		end
		while #bits < 8 do
			bits[#bits + 1] = 0
		end
		local binary = ""
		for index = 1, #bits do
			binary = binary..math.tointeger(bits[#bits - (index - 1)])
		end
		return "0b"..binary
	end
end

--Outputs a string with the given frame data for Assembly
local function assembly_frame(image, last, hex)
	local code = ""
	for y = 1, image.height do
		local chunk_a = 0
		local chunk_b = 0
		for x = 1, image.width do
			local colour = image:getPixel(x - 1, y - 1)
			chunk_a = (chunk_a << 1) | (colour & 1)
			chunk_b = (chunk_b << 1) | ((colour >> 1) & 1)
		end
		code = code.."\t.db "..format_chunk(chunk_a, hex)..", "..format_chunk(chunk_b, hex)
		if not last or y < image.height then code = code.."\n" end
	end
	if not last then code = code.."\n" end
	return code
end

--Outputs a string with the given frame data for C/C++
local function c_frame(image, last, hex)
	local code = ""
	for y = 1, image.height do
		local chunk_a = 0
		local chunk_b = 0
		for x = 1, image.width do
			local colour = image:getPixel(x - 1, y - 1)
			chunk_a = (chunk_a << 1) | (colour & 1)
			chunk_b = (chunk_b << 1) | ((colour >> 1) & 1)
		end
		code = code.."\t"..format_chunk(chunk_a, hex)..", "..format_chunk(chunk_b, hex)
		if not last or y < image.height then code = code..",\n" end
	end
	if not last then code = code.."\n" end
	return code
end

--Writes the pattern details and data to the specified Assembly file
local function write_assembly_file(path, name, sprite, hex)
	local assembly_code = name.."Bank = 0\n"..name.."PatternSize = "..#sprite.frames.."\n"..name.."PatternWidth = "..sprite.width.."\n "..name.."PatternHeight = "..sprite.height.."\n"..name.."PatternPixels:\n"
	for frame_index, frame_info in ipairs(sprite.frames) do
		assembly_code = assembly_code..assembly_frame(frame_info.image, frame_index == #sprite.frames)
		if frame_index < #sprite.frames then assembly_code = assembly_code.."\n" end
	end
	local assembly_file = io.open(path, "w")
	assembly_file:write(assembly_code)
	assembly_file:close()
end

--Writes the pattern details and data to the specified C/C++ files
local function write_c_files(path_header, path_source, name, sprite, hex)
	local header_code = "#ifndef "..name.."Bank\n#define "..name.."Bank 0\n#define "..name.."PatternSize "..#sprite.frames.."\n#define "..name.."PatternWidth "..sprite.width.."\n#define "..name.."PatternHeight "..sprite.height.."\nextern unsigned char "..name.."PatternPixels[];\n#endif"
	local source_code = "unsigned char "..name.."PatternPixels[] = {\n"
	for frame_index, frame_info in ipairs(sprite.frames) do
		source_code = source_code..c_frame(frame_info.image, frame_index == #sprite.frames, hex)
		if frame_index < #sprite.frames then source_code = source_code.."\n" end
	end
	source_code = source_code.."};"
	local header_file = io.open(path_header, "w")
	header_file:write(header_code)
	header_file:close()
	local source_file = io.open(path_source, "w")
	source_file:write(source_code)
	source_file:close()
end

--					SCRIPT					--

--Acquire the needed sprite
local sprite = acquire_sprite()
if sprite == nil then return end

--Create, set up and show the custom dialog then catch the results
local dialog = create_dialog(sprite.filename:gsub("."..app.fs.fileExtension(sprite.filename), ""))
dialog:show{ wait = true }
local dialog_data = dialog.data
if not dialog_data.confirm then return end


--Create and write to the specified export file(s)
local function format_chunk(number, hex)
	if hex then
		return "0x"..string.format("%02x", number);
	else
		local bits = { }
		while number > 0 do
			local current = math.fmod(number, 2)
			bits[#bits + 1] = current
			number = (number - current) / 2
		end
		while #bits < 8 do
			bits[#bits + 1] = 0
		end
		local binary = ""
		for index = 1, #bits do
			binary = binary..math.tointeger(bits[#bits - (index - 1)])
		end
		return "0b"..binary
	end
end
if not dialog_data.assembly then
	local header = io.open(dialog_data.header, "w")
	header:write("#ifndef "..dialog_data.variable.."Bank\n#define "..dialog_data.variable.."Bank 0\n#define "..dialog_data.variable.."PatternSize "..#sprite.frames.."\n#define "..dialog_data.variable.."PatternWidth "..sprite.width.."\n#define "..dialog_data.variable.."PatternHeight "..sprite.height.."\nextern unsigned char "..dialog_data.variable.."PatternPixels[];\n#endif")
	header:close()
end
local code = ""
if dialog_data.assembly then
	code = dialog_data.variable.."Bank = 0\n"..dialog_data.variable.."PatternSize = "..#sprite.frames.."\n"..dialog_data.variable.."PatternWidth = "..sprite.width.."\n"..dialog_data.variable.."PatternHeight = "..sprite.height.."\n"..dialog_data.variable.."PatternPixels:\n"
else
	code = "unsigned char "..dialog_data.variable.."PatternPixels[] = {\n"
end
local image = Image(sprite.width, sprite.height, ColorMode.INDEXED)
for index = 1, #sprite.frames do
	image:clear()
	image:drawSprite(sprite, index)
	for y = 1, sprite.height do
		local chunk_a = 0
		local chunk_b = 0
		for x = 1, sprite.width do
			local colour = image:getPixel(x - 1, y - 1)
			chunk_a = (chunk_a << 1) | (colour & 1)
			chunk_b = (chunk_b << 1) | ((colour >> 1) & 1)
		end
		local line = (dialog_data.assembly and "\t.db " or "\t")..format_chunk(chunk_a, dialog_data.hexadecimal)..", "..format_chunk(chunk_b, dialog_data.hexadecimal)
		code = code..line..((not dialog_data.assembly and (index < #sprite.frames or y < sprite.height)) and ",\n" or "\n")
	end
	if index < #sprite.frames then code = code.."\n" end
end
if dialog_data.assembly then code = string.sub(code, 0, #code - 1)
else code = code.."};" end
local source = io.open((dialog_data.assembly and dialog_data.asm or dialog_data.source), "w")
source:write(code)
source:close()

